
/* ---------------------------------------------------------------------------

    Petit programme d'exemple d'utilisation du port joystick commme
	thermometre.

	Explication complementaire sommaire:

	Le principe du port joystick:

	- un condensateur se charge a travers une resistance raccorde au 5 Volts
      cette resistance est connecte sur le port joystick.

	- quand la tension au borne du condensateur atteint une certaine tension
	  un comparateur change d'etat pour l'indiqu.
      (D'apres des observations faites l'oscsilloscope le comparateur bascule
	    la moiti de la tension d'alimentation environ (soit 2,5 Volts))

	- il suffit donc de mesurer le temp jusqu'au changement d'tat pour avoir
	  une valeur en relation avec la resistance connect (d'aprs la loi t=RC)
	  (t=constante de temps en seconde d'un circuit RC, R=valeur de la
       rsistance en ohms, C=capacit du condensateur en farad)

	- la valeur de resistance utilise pour les joystick se situe generalement
	  autour de 100Kohms

	- un compteur dans une boucle synchronise par l'instruction "IN" permet
	  d'effectuer cela...

	- La valeur de ce compteur a alors directement proportionnel a la resistance.
	  (toujours d'apres t=RC)

	Rmq: Les lecture/ecriture sur le port joystick avec les instructions IN/OUT
	s'effectue a la frequence de 1Mhz (d'apres mon calcul a l'aide du
	programme fp.c (Frequence du Port). Cette frequence etant faible compare
	a celle de fonctionnement des micro-processeurs actuellement les
    instructions IN/OUT joue alors le role de "timer" pour la lecture d'un port.
	La valeur obtenu pour une resistance donne ne devrait pas beaucoup
	changer ( la tolrance des composants prs) d'une machine a l'autre si la
	tension de basculement et la capacit utilis sont les memes.

	Le thermometre:

	- une thermistance de type CTN voit sa rsitance diminu logarithmiquement
      quand la temprature augmente...
	- une

	La temperature peut alors etre interpol par une quation du type suivant:

	T=a*log(R)+b

	pour calcule a et b il suffit d'avoir 2 points de mesure...

	voila c'est tout bete...

	Ce programmme est loin d'etre parfait et propre (temp oblige!!!)
	et pourrait etre ameliore bien sur.
	(En fait le code est vraiment pas tres clean....)
	Mais je laisse a chacun le soin de le personnaliser comme il le desire...
	On pourrait par exemple facilement affich d'autres temprature avec les
	autres entres...

	Remarque:
	 - j'ai consider l'adresse du port joystick fix  0x201
	 il se peut que des ports joysticks soit mapper a d'autres adresses...
	 (voir la routine

	Ce programme a ecrit pour etre compile en 32 bits mais pourrait etre
	compile en 16 bits moyennant quelques modifications mineures.
    (passage des registres en 16 bits dans la routine assembleur principalement)

	Patrick Bricout pour PcTeam le 21/07/97 01:05

--------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <graph.h>
#include <time.h>
#include <i86.h>
#include <math.h>
#include <float.h>
//#include <limits.h>

// =============== Defines claviers ===============

#define KEY_LEFT	75*256
#define KEY_RIGHT	77*256
#define KEY_UP		72*256
#define KEY_DOWN	80*256

#define KEY_FIRST	71*256
#define KEY_LAST	79*256
#define KEY_INS		82*256
#define KEY_DEL		83*256
#define KEY_PGUP	73*256
#define KEY_PGDW	81*256

#define KEY_TAB		9
#define KEY_RETURN	13
#define KEY_CTRL_RETURN	10	/* bizarre c'est le code CTRL-J !!! */
#define KEY_ESC		27

//  =============== Macros ===============

#define NB_ITEM(a)				(sizeof(a)/sizeof(a[0]))

//  =============== Structures ===============

typedef struct
{
	union {
		int raw[2];
		struct {
			int raw1,raw2;	// Valeurs brut du port
		};
	};
	union {
		float temp[2];
		struct {
			float temp1,temp2;	// Temperature en C
		};
	};
} TEMP_CALIBRATION;

typedef struct
{
	int nbItem;
	char **texts;
	int maxWidth;
	int nbEntry;
} MENU;

typedef struct
{
	short r1,c1,r2,c2;
	short w,h;
} TEXT_INFO;

typedef struct
{
	int w,h;
	int oldMode;
	int bOk;
} SCREEN_INFO;

typedef struct
{
	float min,max;
} SCALE;

//  =============== Contex du programme & Variables ===============

int timeScale=100;											// Lecture toutes les 0.1 secondes
TEMP_CALIBRATION tCalibration={803.0,36.0,23.9,100.0};		// Mes points de mesures...
SCALE            tempScale   ={-20.0,110.0};				// Echelle pour le graphique

TEXT_INFO tInfo;

//  =============== Definition du menu ===============

char *mainMenuTexts[]=
{
	"Affichage graphique",
	"Temprature actuelle",
	"Vitesse d'chantillonnage",
	"Calibrage",
	"Sauvegarde des paramtres"
};

MENU mainMenuDef={NB_ITEM(mainMenuTexts),mainMenuTexts};


//  =============== Prototypes ===============

//   Pas de protos ???

//  =============== Quelques fonctions  ===============


void out(unsigned short portAddress,unsigned char data);
#pragma aux out="out dx,al" parm[dx][al];

unsigned char in(unsigned short portAddress);
#pragma aux in="in al,dx" parm[dx] value[al];

void setcolor(int i,int r,int g,int b)
{
	out(0x3c8,i);
	out(0x3c9,r>>2);
	out(0x3c9,g>>2);
	out(0x3c9,b>>2);
}

/* Lit une touche & touche etendue (0+scan_code) */

int getKey(void)
{
	int key=0;

    if (kbhit())
    {
		key = getch();

		if (key==0)
			key = (getch()<<8);

		while (kbhit()) getch();	/* et vide le buffer clavier */
    }

    return key;
};

/* -------------------------------------------------
 *  Routine de lecture du port joystick analogique
 *
 *  Parametres:
 *   - un pointeur sur tableau de 4 long qui
 *     dans lequel seront renvoye les valeurs
 *   - masque de bit correspondant aux entrees que
 *     l'on veut lire (ex: pour lire les entrees
 *     1 et 3 le masque sera 5)
 *
 *  Rmq:
 *   - il est aussi possible d'utiliser le bios
 *
 *  Principe de fonctionnement:
 *   - les interruptions sont coupes pour ne pas
 *     pertuber la boucle de lecture.
 *   - le out decharge tous les condensateurs et
 *     les bits des entrees analogiques passe tous
 *     a 1..
 *   - puis on compte pour chaque une des 4 entrees
 *     jusqu'a quel change d'etat (passe a 0)
 *   - le compteur ECX sert de Garde de fou (TimeOut)
 *
 *  Format du registre joystick:
 *
 *  Bit 7 | Bit 6 | Bit 5 | Bit 4 | Bit 3 | Bit 2 | Bit 1 | Bit 0
 *  ------|-------|-------|-------|-------|-------|-------|------
 *  B2 J2 | B1 J2 | B2 J1 | B1 J1 | J2 Y  | J2 X  | J1 Y  | J1 X
 *
 *  Jn  : Joystick 1 ou 2
 *  Bn  : Bouton 1 ou 2: quand le bit est a zero le bouton est appuye
 *  X/Y : entre analogique correspondant a l'axe X ou Y du joystick
 *
 */

void readJoystick(long *pData,long mask);
#pragma aux readJoystick=\
	"cli"\
	"mov dx,0x201"\
	"out dx,al"\
	"mov ecx,10000"\
	"mov dword ptr[edi],0"\
	"mov dword ptr[edi+4],0"\
	"mov dword ptr[edi+8],0"\
	"mov dword ptr[edi+12],0"\
"poll_loop:"\
	"in	al,dx"\
	"mov ebx,eax"\
	"shr ebx,1"\
	"adc dword ptr[edi],0"\
	"shr ebx,1"\
	"adc dword ptr[edi+4],0"\
	"shr ebx,1"\
	"adc dword ptr[edi+8],0"\
	"shr ebx,1"\
	"adc dword ptr[edi+12],0"\
	"test esi,eax"\
	"loopnz	poll_loop"\
	"sti" parm[edi][esi] modify[eax ebx ecx edx];

/* ---------------------------------------------------
 *  Lecture des entrees numeriques (boutons)
 *  bit 0:bouton 1
 *  bit 1:bouton 2
 *  bit 2:bouton 3
 *  bit 3:bouton 4
 */

int readButtons(void);
#pragma aux readButtons=\
	"xor eax,eax"\
	"mov dx,201h"\
	"in al,dx"\
	"shr eax,4"\
	"xor eax,15"\
	value[eax] modify[eax edx];

/* ---------------------------------------------------------
 *  La meme chose que ci-dessus mais en utilisant le bios.
 *  Attention parfois les fonctions bios ne fonctionnent pas
 */

void biosReadJoystick(long *pData)
{
	union REGS regs;

	regs.h.ah = 0x84;	/* No de fonction */
	regs.w.dx = 1;		/* No de sous fonction */
	int386(0x15,&regs,&regs);
	pData[0]=regs.w.ax;
	pData[1]=regs.w.bx;
	pData[2]=regs.w.cx;
	pData[3]=regs.w.dx;
}

int biosReadButtons(void)
{
	union REGS regs;

	regs.h.ah = 0x84;	/* No de fonction */
	regs.w.dx = 1;		/* No de sous fonction */
	int386(0x15,&regs,&regs);

	return (int)(((unsigned)(regs.h.al)>>4)^15);
}

/* ----------------------------------------------------------------------- */
/*
 * Calcul de la temperature en fonction de la valeur lu sur
 * le port joystick:
 *
 * simple interpolation lineaire dans un domaine semi-log...
 * Remarque:la base du log n'a pas d'importance cela change
 *	        simplement les coefficients...
 */

float convertRawToTemp(float raw)
{
	float r1,r2,t1,t2,a,b;

	// Rmq: les valeurs ci-dessous pourraient etre pre-calcule

	r1= log((float)tCalibration.raw1);
	r2= log((float)tCalibration.raw2);
	t1= tCalibration.temp1;
	t2= tCalibration.temp2;

	a = (t2-t1)/(r2-r1);
	b = (t1*r2-r1*t2)/(r2-r1);

	return a*log(raw)+b;
}

/* ----------------------------------------------------------------------- */

void wait(int mode,clock_t len)
{
	static clock_t startTime;

	if (mode==0)
		startTime=clock();

	if (mode==1)
		while((((clock()-startTime)*1000)/CLOCKS_PER_SEC)<len);
}

/* ----------------------------------------------------------------------- */

// Lecture du port 1 uniquement

int getTimedJoystick(void)
{
	long j[4];
	readJoystick(&j[0],1);
	wait(0,1);			// apparament foire a cause de interruptions coupe dans readjoystick ?
	wait(1,timeScale);

	return j[0];
}

/* ----------------------------------------------------------------------- */


/* ----------------------------------------------- */
/*
 * Quelques fonctions pour interfacer les fonctions
 * graphiques pour une adaptations + facile
 * sur d'autres systemes
 */

int InitVideo(SCREEN_INFO *pInfo)
{
	/* pour faire ca "clean" on pourrait utiliser _getVideoConfig
	 * j'abrege...
	 */
	pInfo->w  = 640;
	pInfo->h  = 480;
	pInfo->bOk= _setvideomode(_VRES256COLOR);
	_setplotaction(_GPSET);

	return pInfo->bOk;
}

void RestoreVideo(SCREEN_INFO *pInfo)
{
	if (pInfo->bOk)
	{
		_setvideomode(_DEFAULTMODE);
	}
}

void SetPlotColor(int color)
{
	_setcolor(color);
}

void Line(int x1,int y1,int x2,int y2)
{
	_moveto(x1,y1);
	_lineto(x2,y2);
}

void Plot(int x,int y)
{
	_setpixel(x,y);
}

void FilledRectangle(int x1,int y1,int x2,int y2)
{
	_rectangle(_GFILLINTERIOR,x1,y1,x2,y2);
}


/* ----------------------------------------------------------------------- */


void drawTemperatureAxises(int w,int h)
{
	int t,y;

	_setlinestyle(0x0303);
	for (t=tempScale.min; t<tempScale.max; t+=10)
	{
		char string[16];
		y=(t-tempScale.min)*h/(tempScale.max-tempScale.min);
		_setcolor(5);
		_moveto(0,h-1-y-4);
		sprintf(string,"%3d",t);
		_outgtext(string);
		if (t!=0)
			_setcolor(6);
		else
			_setcolor(7);
		Line(32,h-1-y,w-1,h-1-y);
	}
	_setlinestyle(0xffff);
}

/* ----------------------------------------------------------------------- */

void plotTemperature(void)
{
	SCREEN_INFO sInfo;

	if (InitVideo(&sInfo))
	{
		int i;
		int xMin=32;
		int x=xMin;
		float temp[1024];
		float oldTemp=10000;

		for (i=0; i<NB_ITEM(temp); temp[i]=10000,i++);

		drawTemperatureAxises(sInfo.w,sInfo.h);

		while (!kbhit())
		{
			int v=getTimedJoystick();
			if (v!=10000)
				temp[x]=convertRawToTemp(v);
			else
				temp[x]=10000;

			_setcolor(0);
			if (v!=10000 && oldTemp!=10000)
			{
				_setcolor(1);
				Line(x,sInfo.h-1-sInfo.h*(oldTemp-tempScale.min)/(tempScale.max-tempScale.min),
				     x,sInfo.h-1-sInfo.h*(temp[x]-tempScale.min)/(tempScale.max-tempScale.min));
				//Plot(x,sInfo.h-1-sInfo.h*(temp[x]-tempScale.min)/(tempScale.max-tempScale.min));
			}
			oldTemp=temp[x];
			x++;
			// Gestion pseudo historique
			if (x==sInfo.w)
			{
				_clearscreen(_GCLEARSCREEN);
				drawTemperatureAxises(sInfo.w,sInfo.h);
				_setcolor(1);
				for (x=xMin; x<sInfo.w/2; x++)
				{
					temp[x]=temp[x+sInfo.w/2-xMin];
					if (temp[x]!=10000 && temp[x-1]!=10000)
					{
						Line(x,sInfo.h-1-sInfo.h*(temp[x-1]-tempScale.min)/(tempScale.max-tempScale.min),
						     x,sInfo.h-1-sInfo.h*(temp[x]-tempScale.min)/(tempScale.max-tempScale.min));
						//Plot(x,sInfo.h-1-sInfo.h*(temp[x]-tempScale.min)/(tempScale.max-tempScale.min));
					}
				}
				x=sInfo.w/2;
			}
		}
		getch();
		RestoreVideo(&sInfo);
	}
	else
		printf("error setvideomode %08x\n",(int)_grstatus());
}

/* ----------------------------------------------------------------------- */


void printInput(void)
{
	while (!kbhit())
	{
		long j[4];
		readJoystick(&j[0],15);
		printf("%5d %5d %5d %5d\n",j[0],j[1],j[2],j[3]);

	}
	getch();
}


/* ----------------------------------------------------------------------- */

void printTemperature(void)
{
	printf("\n\nValeur actuelle taper une touche pour le retour au menu\n");
	while (!kbhit())
	{
		int v=getTimedJoystick();
		if (v==10000)
			printf("Port sans connection...\n");
		else
			printf("Valeur brute:%5d  Temprature:%3.1f\r",v,convertRawToTemp(v));
	}
	printf("\n");
	getch();
}

/* ----------------------------------------------------------------------- */

void DisplayContext(void)
{
	printf("\n\nConfiguration actuelle:\n"
		   "Valeurs de rfrence 1:%d => temprature en C:%3.1f\n"
		   "Valeurs de rfrence 2:%d => temprature en C:%3.1f\n"
		   "Vitesse d'echantillonnage en ms:%d\n"
			,tCalibration.raw1,tCalibration.temp1,tCalibration.raw2,tCalibration.temp2
			,timeScale
		   );
}

/* ----------------------------------------------------------------------- */

void InputTimeScale(void)
{
	printf("\n\n"
		   "Entrer la vitesse d'echantillonage en ms pour le graphique:\n");
	scanf("%d",&timeScale);
	printf("Valeur lu:%d\n",timeScale);
	printf("Taper une touche pour le retour au menu....\n");
	getch();
}


/* ----------------------------------------------------------------------- */

int DisplayRawValue(void)
{
	int v;

	while (!kbhit())
	{
		v = getTimedJoystick();
		printf("Valeur du port:%d\r",v);
	}
	printf("\n");
	getch();

	return v;
}

/* ----------------------------------------------------------------------- */

void InputCalibration(void)
{
	int i;

	_clearscreen(_GCLEARSCREEN);
	printf("Calibrage de la thermistance:\n"
		   "La calibrage necessite 2 valeurs de rfrence:\n"
		   "Attendez une valeur approximativement stable.\n"
		   "Tapez ESC pour retourner au menu tout de suite.\n"
		   );
	if (getch()!=KEY_ESC)
	{
		for (i=0; i<2; i++)
		{
			printf("La valeur du port va s'afficher taper une touche\n"
		       	"pour entrer la %s temprature\n",i?"1re":"2me"
			  	);

			tCalibration.raw[i] = DisplayRawValue();

			printf("Entrer la temperature en C correspondante sous la forme xx.x:\n");
			scanf("%f",&tCalibration.temp[i]);
			printf("Valeur lu: %3.1fC\n",tCalibration.temp[i]);
		}
		printf("Taper une touche pour le retour au menu....\n");
		getch();
	}
}

/* ----------------------------------------------------------------
 *  Une petite routine de menu
 */

void CenterString(char *string,int y)
{
	_settextposition(y,(tInfo.w-strlen(string))/2);
	_outtext(string);
}

//_displaycursor(_GCURSORON);

void InitMenu(MENU *pMenu)
{
	int i;

	pMenu->maxWidth=0;
	for (i=0; i<pMenu->nbItem; i++)
	{
		pMenu->maxWidth=max(pMenu->maxWidth,strlen(pMenu->texts[i]));
	}
	pMenu->nbEntry=pMenu->nbItem;
}


void DrawMenu(MENU *pMenu,int selected,int y)
{
	int i;
	int x;
	short oldTextColor=_settextcolor(_BLACK);
	short oldBackColor=_setbkcolor(_BLACK);

	x=(tInfo.w-pMenu->maxWidth)/2;

	for (i=0; i<pMenu->nbItem; i++)
	{
		_settextposition(y++,x);
		if (i==selected)
		{
			_setbkcolor(_BLUE);
			_settextcolor(15);
		}
		else
		{
			_setbkcolor(_BLACK);
			_settextcolor(7);
		}
		_outtext(pMenu->texts[i]);
	}
	_settextcolor(oldTextColor);
	_setbkcolor(oldBackColor);
}



int Menu(MENU *pMenu)
{
	int bExit=0;
	int choice=0;

	InitMenu(pMenu);

	// Lecture des caracteriques de l'ecran
	_gettextwindow(&tInfo.r1,&tInfo.c1,&tInfo.r2,&tInfo.c2);	// valeurs typique:1,1,50,80
	tInfo.w = tInfo.c2-tInfo.c1+1;
	tInfo.h = tInfo.r2-tInfo.r1+1;

	// efface l'ecran
	_clearscreen(_GCLEARSCREEN);
	CenterString("TeamBidouille TermoJoystick",3);

	while (!bExit)
	{
		int c;

		DrawMenu(pMenu,choice,10);
		DisplayContext();

		while (!kbhit());
		c = getKey();
		switch(c)
		{
			case KEY_UP:
				if (choice>0)
					choice--;
				break;
			case KEY_DOWN:
				if (choice<pMenu->nbEntry-1)
					choice++;
				break;
			case KEY_RETURN:
				bExit=1;
				break;
			case KEY_ESC:
				bExit=1;
				choice=-1;
				break;
		}
	}

	return choice;
}

/* ---------------------------------------------------------------------- */

void ReadContext(void)
{
	FILE *file;

	if (file=fopen("termo.cnf","r"))
	{
		fscanf(file,"Calibration:%d %f %d %f\n",
		       &tCalibration.raw1,&tCalibration.temp1,&tCalibration.raw2,&tCalibration.temp2);
		fscanf(file,"TimeScale:%d ms\n",&timeScale);
		fscanf(file,"TempScale:%f %f\n",&tempScale.min,&tempScale.max);
		fclose(file);
	}
}

/* ---------------------------------------------------------------------- */

void WriteContext(void)
{
	FILE *file;

	if (file=fopen("termo.cnf","w"))
	{
		fprintf(file,"Calibration:%d %f %d %f\n",
		       tCalibration.raw1,tCalibration.temp1,tCalibration.raw2,tCalibration.temp2);
		fprintf(file,"TimeScale:%d ms\n",timeScale);
		fprintf(file,"TempScale:%f %f\n",tempScale.min,tempScale.max);
		fclose(file);
	}
}


/* ----------------------------------------------------------------------- */

void main(void)
{
	int choice;

	ReadContext();

	do
	{
		choice = Menu(&mainMenuDef);
		switch (choice)
		{
			case 0:
				plotTemperature();
				break;
			case 1:
				printTemperature();
				break;
			case 2:
				InputTimeScale();
				break;
			case 3:
				InputCalibration();
				break;
			case 4:
				WriteContext();
				break;
		}
	}
	while (choice!=-1);
}
